from typing import List, Union

import torch
from transformers import AutoModelForCausalLM, AutoTokenizer, PreTrainedModel
from transformers.tokenization_utils_base import BatchEncoding


def rank(
    text: str,
    model_ids: List[str],
    stride: int = 512,
    device: Union[str, torch.device] = "cpu",
):
    # TODO: parallelize
    # TODO: support batches
    model2ppl = {}
    for model_id in model_ids:
        model = AutoModelForCausalLM.from_pretrained(model_id)
        tokenizer = AutoTokenizer.from_pretrained(model_id)
        encodings = tokenizer(text, return_tensors="pt")
        perplexity = get_perplexity(model, encodings, stride, device)
        model2ppl[model_id] = perplexity
    return {
        model_id: perplexity
        for model_id, perplexity in sorted(model2ppl.items(), reverse=True)
    }


@torch.inference_mode()
def get_perplexity(
    model: PreTrainedModel,
    encodings: BatchEncoding,
    stride: int,
    device: Union[str, torch.device] = "cpu",
) -> float:
    # from https://huggingface.co/docs/transformers/perplexity
    nlls = []
    prev_end_loc = 0
    max_length = model.config.n_positions
    seq_len = encodings.input_ids.size(1)
    model.eval()
    model.to(device)
    for begin_loc in range(0, seq_len, stride):
        end_loc = min(begin_loc + max_length, seq_len)
        trg_len = end_loc - prev_end_loc
        input_ids = encodings.input_ids[:, begin_loc:end_loc].to(device)
        target_ids = input_ids.clone()
        target_ids[:, :-trg_len] = -100
        outputs = model(input_ids, labels=target_ids)
        neg_log_likelihood = outputs.loss * trg_len
        nlls.append(neg_log_likelihood)
        prev_end_loc = end_loc
        if end_loc == seq_len:
            break
    ppl = torch.exp(torch.stack(nlls).sum() / end_loc).item()
    return ppl



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['lm_identifier']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.24.1,<2.0.0', 'torch>=1.13.1,<2.0.0', 'transformers>=4.25.1,<5.0.0']

setup_kwargs = {
    'name': 'lm-identifier',
    'version': '0.0.1',
    'description': 'A toolkit for identifying pretrained language models from AI-generated text',
    'long_description': '# LM Identifier\n\nWith a surge of generative pretrained language models, it is becoming increasingly important to distinguish between human and AI-generated text. Inspired by [GPTZero](https://etedward-gptzero-main-zqgfwb.streamlit.app), an app that seeks to detect AI-generated text, LM Identifier pokes at this question even further by providing a growing suite of tools to help identify *which (publicly available) language model* might have been used to generate some given chunck of text.\n\n## Installation\n\nLM Identifier is available on PyPI.\n\n```\n$ pip install lm-identifier\n```\n\nTo develop locally, first install pre-commit:\n\n```\n$ pip install --upgrade pip wheel\n$ pip install pre-commit\n$ pre-commit install\n```\n\nInstall the package in editable mode.\n\n```\npip install -e .\n```\n\n## Usages\n\nWIP\n',
    'author': 'Jaesung Tae',
    'author_email': 'jaesungtae@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
